@@ -1,17 +1,9 @@
 // Take a look at the license at the top of the repository in the LICENSE file.
 
-use glib::object::IsA;
-use glib::translate::*;
-use glib::Error;
-use libc::{c_uchar, c_void};
-use std::io::Read;
-use std::mem;
-use std::path::Path;
-use std::pin::Pin;
-use std::ptr;
-use std::slice;
+use std::{future::Future, io::Read, mem, path::Path, pin::Pin, ptr, slice};
 
-use std::future::Future;
+use glib::{prelude::*, translate::*, Error};
+use libc::{c_uchar, c_void};
 
 use crate::{Colorspace, Pixbuf, PixbufFormat};
 
@@ -32,8 +24,8 @@ impl Pixbuf {
         assert!(width > 0, "width must be greater than 0");
         assert!(height > 0, "height must be greater than 0");
         assert!(row_stride > 0, "row_stride must be greater than 0");
-        assert!(
-            bits_per_sample == 8,
+        assert_eq!(
+            bits_per_sample, 8,
             "bits_per_sample == 8 is the only supported value"
         );
 
@@ -50,7 +42,7 @@ impl Pixbuf {
         let ptr = {
             let data: &mut [u8] = (*data).as_mut();
             assert!(
-                data.len() >= ((height - 1) * row_stride + last_row_len) as usize,
+                data.len() >= ((height - 1) * row_stride + last_row_len),
                 "data.len() must fit the width, height, and row_stride"
             );
             data.as_mut_ptr()
@@ -85,97 +77,31 @@ impl Pixbuf {
         Pixbuf::from_stream(&gio::ReadInputStream::new(r), None::<&gio::Cancellable>)
     }
 
-    #[doc(alias = "gdk_pixbuf_new_from_file")]
-    #[doc(alias = "gdk_pixbuf_new_from_file_utf8")]
-    pub fn from_file<T: AsRef<Path>>(filename: T) -> Result<Pixbuf, Error> {
-        #[cfg(not(windows))]
-        use ffi::gdk_pixbuf_new_from_file;
-        #[cfg(windows)]
-        use ffi::gdk_pixbuf_new_from_file_utf8 as gdk_pixbuf_new_from_file;
-
-        unsafe {
-            let mut error = ptr::null_mut();
-            let ptr = gdk_pixbuf_new_from_file(filename.as_ref().to_glib_none().0, &mut error);
-            if error.is_null() {
-                Ok(from_glib_full(ptr))
-            } else {
-                Err(from_glib_full(error))
-            }
-        }
-    }
-
-    #[doc(alias = "gdk_pixbuf_new_from_file_at_size")]
-    #[doc(alias = "gdk_pixbuf_new_from_file_at_size_utf8")]
-    pub fn from_file_at_size<T: AsRef<Path>>(
-        filename: T,
-        width: i32,
-        height: i32,
-    ) -> Result<Pixbuf, Error> {
-        #[cfg(not(windows))]
-        use ffi::gdk_pixbuf_new_from_file_at_size;
-        #[cfg(windows)]
-        use ffi::gdk_pixbuf_new_from_file_at_size_utf8 as gdk_pixbuf_new_from_file_at_size;
-
-        unsafe {
-            let mut error = ptr::null_mut();
-            let ptr = gdk_pixbuf_new_from_file_at_size(
-                filename.as_ref().to_glib_none().0,
-                width,
-                height,
-                &mut error,
-            );
-            if error.is_null() {
-                Ok(from_glib_full(ptr))
-            } else {
-                Err(from_glib_full(error))
-            }
-        }
-    }
-
-    #[doc(alias = "gdk_pixbuf_new_from_file_at_scale")]
-    #[doc(alias = "gdk_pixbuf_new_from_file_at_scale_utf8")]
-    pub fn from_file_at_scale<T: AsRef<Path>>(
-        filename: T,
-        width: i32,
-        height: i32,
-        preserve_aspect_ratio: bool,
-    ) -> Result<Pixbuf, Error> {
-        #[cfg(not(windows))]
-        use ffi::gdk_pixbuf_new_from_file_at_scale;
-        #[cfg(windows)]
-        use ffi::gdk_pixbuf_new_from_file_at_scale_utf8 as gdk_pixbuf_new_from_file_at_scale;
-
-        unsafe {
-            let mut error = ptr::null_mut();
-            let ptr = gdk_pixbuf_new_from_file_at_scale(
-                filename.as_ref().to_glib_none().0,
-                width,
-                height,
-                preserve_aspect_ratio.into_glib(),
-                &mut error,
-            );
-            if error.is_null() {
-                Ok(from_glib_full(ptr))
-            } else {
-                Err(from_glib_full(error))
-            }
-        }
-    }
-
     #[doc(alias = "gdk_pixbuf_new_from_stream_async")]
     pub fn from_stream_async<
         P: IsA<gio::InputStream>,
         Q: IsA<gio::Cancellable>,
-        R: FnOnce(Result<Pixbuf, Error>) + Send + 'static,
+        R: FnOnce(Result<Pixbuf, Error>) + 'static,
     >(
         stream: &P,
         cancellable: Option<&Q>,
         callback: R,
     ) {
+        let main_context = glib::MainContext::ref_thread_default();
+        let is_main_context_owner = main_context.is_owner();
+        let has_acquired_main_context = (!is_main_context_owner)
+            .then(|| main_context.acquire().ok())
+            .flatten();
+        assert!(
+            is_main_context_owner || has_acquired_main_context.is_some(),
+            "Async operations only allowed if the thread is owning the MainContext"
+        );
+
         let cancellable = cancellable.map(|p| p.as_ref());
-        let user_data: Box<R> = Box::new(callback);
+        let user_data: Box<glib::thread_guard::ThreadGuard<R>> =
+            Box::new(glib::thread_guard::ThreadGuard::new(callback));
         unsafe extern "C" fn from_stream_async_trampoline<
-            R: FnOnce(Result<Pixbuf, Error>) + Send + 'static,
+            R: FnOnce(Result<Pixbuf, Error>) + 'static,
         >(
             _source_object: *mut glib::gobject_ffi::GObject,
             res: *mut gio::ffi::GAsyncResult,
@@ -188,7 +114,9 @@ impl Pixbuf {
             } else {
                 Err(from_glib_full(error))
             };
-            let callback: Box<R> = Box::from_raw(user_data as *mut _);
+            let callback: Box<glib::thread_guard::ThreadGuard<R>> =
+                Box::from_raw(user_data as *mut _);
+            let callback = callback.into_inner();
             callback(result);
         }
         let callback = from_stream_async_trampoline::<R>;
@@ -202,7 +130,7 @@ impl Pixbuf {
         }
     }
 
-    pub fn from_stream_async_future<P: IsA<gio::InputStream> + Clone + 'static>(
+    pub fn from_stream_future<P: IsA<gio::InputStream> + Clone + 'static>(
         stream: &P,
     ) -> Pin<Box<dyn Future<Output = Result<Pixbuf, Error>> + 'static>> {
         let stream = stream.clone();
@@ -217,7 +145,7 @@ impl Pixbuf {
     pub fn from_stream_at_scale_async<
         P: IsA<gio::InputStream>,
         Q: IsA<gio::Cancellable>,
-        R: FnOnce(Result<Pixbuf, Error>) + Send + 'static,
+        R: FnOnce(Result<Pixbuf, Error>) + 'static,
     >(
         stream: &P,
         width: i32,
@@ -226,10 +154,21 @@ impl Pixbuf {
         cancellable: Option<&Q>,
         callback: R,
     ) {
+        let main_context = glib::MainContext::ref_thread_default();
+        let is_main_context_owner = main_context.is_owner();
+        let has_acquired_main_context = (!is_main_context_owner)
+            .then(|| main_context.acquire().ok())
+            .flatten();
+        assert!(
+            is_main_context_owner || has_acquired_main_context.is_some(),
+            "Async operations only allowed if the thread is owning the MainContext"
+        );
+
         let cancellable = cancellable.map(|p| p.as_ref());
-        let user_data: Box<R> = Box::new(callback);
+        let user_data: Box<glib::thread_guard::ThreadGuard<R>> =
+            Box::new(glib::thread_guard::ThreadGuard::new(callback));
         unsafe extern "C" fn from_stream_at_scale_async_trampoline<
-            R: FnOnce(Result<Pixbuf, Error>) + Send + 'static,
+            R: FnOnce(Result<Pixbuf, Error>) + 'static,
         >(
             _source_object: *mut glib::gobject_ffi::GObject,
             res: *mut gio::ffi::GAsyncResult,
@@ -242,7 +181,9 @@ impl Pixbuf {
             } else {
                 Err(from_glib_full(error))
             };
-            let callback: Box<R> = Box::from_raw(user_data as *mut _);
+            let callback: Box<glib::thread_guard::ThreadGuard<R>> =
+                Box::from_raw(user_data as *mut _);
+            let callback = callback.into_inner();
             callback(result);
         }
         let callback = from_stream_at_scale_async_trampoline::<R>;
@@ -259,7 +200,7 @@ impl Pixbuf {
         }
     }
 
-    pub fn from_stream_at_scale_async_future<P: IsA<gio::InputStream> + Clone + 'static>(
+    pub fn from_stream_at_scale_future<P: IsA<gio::InputStream> + Clone + 'static>(
         stream: &P,
         width: i32,
         height: i32,
@@ -280,6 +221,19 @@ impl Pixbuf {
         }))
     }
 
+    // rustdoc-stripper-ignore-next
+    /// Returns a mutable slice to the pixbuf's pixel data.
+    ///
+    /// This function will cause an implicit copy if the pixbuf was created from read-only data.
+    ///
+    /// Please see the section on [image data](#image-data) for information about how the pixel
+    /// data is stored in memory.
+    ///
+    /// # Safety
+    /// No other reference to this pixbuf's data must exist when this method is called.
+    ///
+    /// Until you drop the returned reference, you must not call any methods on the pixbuf which may read
+    /// or write to the data.
     #[allow(clippy::mut_from_ref)]
     #[allow(clippy::missing_safety_doc)]
     #[doc(alias = "gdk_pixbuf_get_pixels_with_length")]
@@ -287,6 +241,9 @@ impl Pixbuf {
     pub unsafe fn pixels(&self) -> &mut [u8] {
         let mut len = 0;
         let ptr = ffi::gdk_pixbuf_get_pixels_with_length(self.to_glib_none().0, &mut len);
+        if len == 0 {
+            return &mut [];
+        }
         slice::from_raw_parts_mut(ptr, len as usize)
     }
 
@@ -345,17 +302,28 @@ impl Pixbuf {
     #[doc(alias = "get_file_info_async")]
     pub fn file_info_async<
         P: IsA<gio::Cancellable>,
-        Q: FnOnce(Result<Option<(PixbufFormat, i32, i32)>, Error>) + Send + 'static,
+        Q: FnOnce(Result<Option<(PixbufFormat, i32, i32)>, Error>) + 'static,
         T: AsRef<Path>,
     >(
         filename: T,
         cancellable: Option<&P>,
         callback: Q,
     ) {
+        let main_context = glib::MainContext::ref_thread_default();
+        let is_main_context_owner = main_context.is_owner();
+        let has_acquired_main_context = (!is_main_context_owner)
+            .then(|| main_context.acquire().ok())
+            .flatten();
+        assert!(
+            is_main_context_owner || has_acquired_main_context.is_some(),
+            "Async operations only allowed if the thread is owning the MainContext"
+        );
+
         let cancellable = cancellable.map(|p| p.as_ref());
-        let user_data: Box<Q> = Box::new(callback);
+        let user_data: Box<glib::thread_guard::ThreadGuard<Q>> =
+            Box::new(glib::thread_guard::ThreadGuard::new(callback));
         unsafe extern "C" fn get_file_info_async_trampoline<
-            Q: FnOnce(Result<Option<(PixbufFormat, i32, i32)>, Error>) + Send + 'static,
+            Q: FnOnce(Result<Option<(PixbufFormat, i32, i32)>, Error>) + 'static,
         >(
             _source_object: *mut glib::gobject_ffi::GObject,
             res: *mut gio::ffi::GAsyncResult,
@@ -381,7 +349,9 @@ impl Pixbuf {
                     height.assume_init(),
                 )))
             };
-            let callback: Box<Q> = Box::from_raw(user_data as *mut _);
+            let callback: Box<glib::thread_guard::ThreadGuard<Q>> =
+                Box::from_raw(user_data as *mut _);
+            let callback = callback.into_inner();
             callback(result);
         }
         let callback = get_file_info_async_trampoline::<Q>;
@@ -396,8 +366,8 @@ impl Pixbuf {
     }
 
     #[allow(clippy::type_complexity)]
-    #[doc(alias = "get_file_info_async_future")]
-    pub fn file_info_async_future<T: AsRef<Path> + Clone + 'static>(
+    #[doc(alias = "get_file_info_async")]
+    pub fn file_info_future<T: AsRef<Path> + Clone + 'static>(
         filename: T,
     ) -> Pin<Box<dyn Future<Output = Result<Option<(PixbufFormat, i32, i32)>, Error>> + 'static>>
     {
@@ -428,7 +398,7 @@ impl Pixbuf {
             if error.is_null() {
                 Ok(FromGlibContainer::from_glib_full_num(
                     buffer,
-                    buffer_size.assume_init() as usize,
+                    buffer_size.assume_init() as _,
                 ))
             } else {
                 Err(from_glib_full(error))
@@ -436,8 +406,6 @@ impl Pixbuf {
         }
     }
 
-    #[cfg(any(feature = "v2_36", feature = "dox"))]
-    #[cfg_attr(feature = "dox", doc(cfg(feature = "v2_36")))]
     #[doc(alias = "gdk_pixbuf_save_to_streamv")]
     pub fn save_to_streamv<P: IsA<gio::OutputStream>, Q: IsA<gio::Cancellable>>(
         &self,
@@ -468,13 +436,11 @@ impl Pixbuf {
         }
     }
 
-    #[cfg(any(feature = "v2_36", feature = "dox"))]
-    #[cfg_attr(feature = "dox", doc(cfg(feature = "v2_36")))]
     #[doc(alias = "gdk_pixbuf_save_to_streamv_async")]
     pub fn save_to_streamv_async<
         P: IsA<gio::OutputStream>,
         Q: IsA<gio::Cancellable>,
-        R: FnOnce(Result<(), Error>) + Send + 'static,
+        R: FnOnce(Result<(), Error>) + 'static,
     >(
         &self,
         stream: &P,
@@ -483,10 +449,21 @@ impl Pixbuf {
         cancellable: Option<&Q>,
         callback: R,
     ) {
+        let main_context = glib::MainContext::ref_thread_default();
+        let is_main_context_owner = main_context.is_owner();
+        let has_acquired_main_context = (!is_main_context_owner)
+            .then(|| main_context.acquire().ok())
+            .flatten();
+        assert!(
+            is_main_context_owner || has_acquired_main_context.is_some(),
+            "Async operations only allowed if the thread is owning the MainContext"
+        );
+
         let cancellable = cancellable.map(|p| p.as_ref());
-        let user_data: Box<R> = Box::new(callback);
+        let user_data: Box<glib::thread_guard::ThreadGuard<R>> =
+            Box::new(glib::thread_guard::ThreadGuard::new(callback));
         unsafe extern "C" fn save_to_streamv_async_trampoline<
-            R: FnOnce(Result<(), Error>) + Send + 'static,
+            R: FnOnce(Result<(), Error>) + 'static,
         >(
             _source_object: *mut glib::gobject_ffi::GObject,
             res: *mut gio::ffi::GAsyncResult,
@@ -499,7 +476,9 @@ impl Pixbuf {
             } else {
                 Err(from_glib_full(error))
             };
-            let callback: Box<R> = Box::from_raw(user_data as *mut _);
+            let callback: Box<glib::thread_guard::ThreadGuard<R>> =
+                Box::from_raw(user_data as *mut _);
+            let callback = callback.into_inner();
             callback(result);
         }
         let callback = save_to_streamv_async_trampoline::<R>;
@@ -519,9 +498,7 @@ impl Pixbuf {
         }
     }
 
-    #[cfg(any(feature = "v2_36", feature = "dox"))]
-    #[cfg_attr(feature = "dox", doc(cfg(feature = "v2_36")))]
-    pub fn save_to_streamv_async_future<P: IsA<gio::OutputStream> + Clone + 'static>(
+    pub fn save_to_streamv_future<P: IsA<gio::OutputStream> + Clone + 'static>(
         &self,
         stream: &P,
         type_: &str,
@@ -536,7 +513,7 @@ impl Pixbuf {
         Box::pin(gio::GioFuture::new(self, move |obj, cancellable, send| {
             let options = options
                 .iter()
-                .map(|&(ref k, ref v)| (k.as_str(), v.as_str()))
+                .map(|(k, v)| (k.as_str(), v.as_str()))
                 .collect::<Vec<(&str, &str)>>();
 
             obj.save_to_streamv_async(
